/*
 * Kernel Module Wrapper
 * Copyright (C) 2001 Marco Lamberto <lm@sunnyspot.org>
 * Web page: http://kmw.sourceforge.net/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id$
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/init.h>
#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif
#include "kmw.h"


MODULE_AUTHOR("Marco Lamberto <lm@sunnyspot.org>");
MODULE_DESCRIPTION("Linux Kernel Module Wrapper " KMW_VERSION);


typedef struct {
	int				free : 1;													/* free entry */
	char			modname[KMW_MAX_MODNAME + 1];			/* module name */
	kmw_hook	*hooks;														/* hooks list */
#ifdef CONFIG_PROC_FS
	struct proc_dir_entry dir;									/* proc fs directory entry */
#endif
} kmw_mod_list;

#define KMW_HOOK_LIST_END(list) (list.f_org == NULL || list.f_new == NULL)

kmw_mod_list kmw_module_list[KMW_MAX_MODULES];


int init_module(void);
void cleanup_module(void);
int kmw_toggle_hook(kmw_hook *hook, int status);
int kmw_init_hook(kmw_modid modid, kmw_hook *hook);
int kmw_remove_hook(kmw_modid modid, kmw_hook *hook);
kmw_modid kmw_register_hooks(char *modname, kmw_hook *h_list);
int kmw_unregister_hooks(kmw_modid modid);


/*
EXPORT_SYMBOL(kmw_register_hooks);
EXPORT_SYMBOL(kmw_unregister_hooks);
*/



#ifdef CONFIG_PROC_FS
struct proc_dir_entry proc_kmw =
{
	0,
	3, "kmw",
	S_IFDIR | S_IRUGO | S_IXUGO,
	1, 0, 0, 0,
	NULL,		/* functions that can be done on the inode */
	NULL,		/* read */
	NULL
};
#endif

	
	
int
init_module()
{
	int i;
	
	printk(KERN_INFO "KMW %s loaded\n", KMW_VERSION);
	for (i = 0; i < KMW_MAX_MODULES; i++) {
		kmw_module_list[i].free = TRUE;
	}
#ifdef CONFIG_PROC_FS
	return proc_register(&proc_root, &proc_kmw);
#else
	return FALSE;
#endif
}

void
cleanup_module()
{
#ifdef CONFIG_PROC_FS
	proc_unregister(&proc_root, proc_kmw.low_ino);
#endif
	printk(KERN_INFO "KMW unloaded\n");
}

int
kmw_toggle_hook(kmw_hook *hook, int status)
{
	if (hook->init && hook->active != status) {
		printk(KERN_INFO "KMW toggling hook '%s':%d->%d\n",
			hook->name, hook->active, status);
		hook->active = status;
		/* FIXME: is atomic ?? */
		memcpy(hook->f_org,
			status ? hook->b_new : hook->b_org,
			sizeof(hook->b_new));
	}
	return FALSE;
}

int
kmw_init_hook(kmw_modid modid, kmw_hook *hook)
{
	printk(KERN_INFO "KMW init hook '%s':%p->%p\n",
		hook->name, hook->f_org, hook->f_new);

	hook->init = FALSE;
	hook->active = FALSE;

	/* FIXME: the jmp *addr opcode is machine dep! */
	hook->b_new[0] = 0xff;
	hook->b_new[1] = 0x25;
	*(int *)(hook->b_new + 2) = (int)&hook->f_new;

	/* check if function was already hooked */
	/* FIXME: the jmp opcode length is machine dep! */
	if (memcmp(hook->f_org, hook->b_new, 2) == 0) {
		printk(KERN_INFO "KMW '%s' was already hooked\n", hook->name);
		return TRUE;
	}
	
	/* save the original entry point */
	memcpy(hook->b_org, hook->f_org, sizeof(hook->b_org));

	hook->init = TRUE;

	/* activate on load */
	kmw_toggle_hook(hook, hook->activate);
	
#ifdef CONFIG_PROC_FS
	hook->inode_ops.default_file_ops = &hook->file_ops;
	hook->file.ops = &hook->inode_ops;
	hook->file.namelen = strlen(hook->file.name);
	*hook->self = hook;
	proc_register(&kmw_module_list[modid].dir, &hook->file);
#endif

	return FALSE;
}

int
kmw_remove_hook(kmw_modid modid, kmw_hook *hook)
{
	kmw_toggle_hook(hook, FALSE);

#ifdef CONFIG_PROC_FS
	proc_unregister(&kmw_module_list[modid].dir, hook->file.low_ino);
#endif

	return FALSE;
}

kmw_modid
kmw_register_hooks(char *modname, kmw_hook *h_list)
{
	int i;
	int pos = -1;

	/* find first free slot */
	for (i = 0; i < KMW_MAX_MODULES; i++) {
		/* FIXME: sync problems: is it safe/atomic?? */
		if (kmw_module_list[i].free == TRUE) {
			kmw_module_list[i].free = FALSE;
			pos = i;
			break;
		}
	}
	if (pos < 0) {
		printk(KERN_INFO "KMW cannot register another module\n");
		return TRUE;
	}

#ifdef CONFIG_PROC_FS
	printk(KERN_INFO "KMW registering /proc/kmw/%s\n", modname);
	memset(&kmw_module_list[pos].dir, 0, sizeof(struct proc_dir_entry));
	kmw_module_list[pos].dir.namelen = strlen(modname);
	kmw_module_list[pos].dir.name = modname;
	kmw_module_list[pos].dir.mode = S_IFDIR | S_IRUGO | S_IXUGO;
	proc_register(&proc_kmw, &kmw_module_list[pos].dir);
	printk(KERN_INFO "KMW registered /proc/kmw/%s\n", modname);
#endif

	printk(KERN_INFO "KMW registering hooks '%s':%d\n", modname, pos);
	strncpy(kmw_module_list[pos].modname, modname, KMW_MAX_MODNAME);
	kmw_module_list[pos].hooks = h_list;
	for (i = 0; !KMW_HOOK_LIST_END(h_list[i]); i++)
		kmw_init_hook(pos, &h_list[i]);
	printk(KERN_INFO "KMW registered hooks '%s':%d\n", modname, pos);

	return pos;
}

int
kmw_unregister_hooks(kmw_modid modid)
{
	int i;
	kmw_hook *h_list = kmw_module_list[modid].hooks;

#ifdef CONFIG_PROC_FS
	printk(KERN_INFO "KMW unregistering /proc/kmw/%s\n",
		kmw_module_list[modid].modname);
	proc_unregister(&proc_kmw, kmw_module_list[modid].dir.low_ino);
	printk(KERN_INFO "KMW unregistered /proc/kmw/%s\n",
		kmw_module_list[modid].modname);
#endif

	printk(KERN_INFO "KMW unregistering hooks '%s':%d\n",
		kmw_module_list[modid].modname, modid);
	for (i = 0; !KMW_HOOK_LIST_END(h_list[i]); i++)
		kmw_remove_hook(modid, &h_list[i]);
	kmw_module_list[modid].free = TRUE;
	printk(KERN_INFO "KMW unregistered hooks '%s':%d\n",
		kmw_module_list[modid].modname, modid);

	return FALSE;
}

/* vim: set ts=2: */
